-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THAbstractSystem = {}
local THAbstractSystem_mt = THUtils.createClass(THAbstractSystem)
THAbstractSystem.networkObjects = {}
THAbstractSystem.networkObjectToId = {}
THAbstractSystem.NETWORK_ID_BITS = 8
THAbstractSystem.MESSAGE = {
    NETWORK_ID_MISMATCH = "Network id mismatch (%s)",
    NETWORK_ID_LIMIT_REACHED = "Maximum network objects (%s) exceeded"
}
function THAbstractSystem.initXMLSchema(xmlSchema, xmlPath)
end
function THAbstractSystem.initXMLSchemaSettings(xmlSchema, xmlPath)
end
function THAbstractSystem.initXMLSchemaSavegame(xmlSchema, xmlPath)
end
function THAbstractSystem.new(rootEnv, dataKey, modName, customMt)
    customMt = customMt or THAbstractSystem_mt
    if rootEnv ~= nil then
        if not THUtils.getIsType(rootEnv, THCore) then
            THUtils.stackMsg("THCore class requred for THAbstractSystem rootEnv")
            return
        end
        dataKey = dataKey or rootEnv.dataKey
        modName = modName or rootEnv.modName
    end
    if THUtils.argIsValid(type(dataKey) == "string" and dataKey ~= "", "dataKey", dataKey)
        and THUtils.argIsValid(type(modName) == "string", "modName", modName)
        and THUtils.argIsValid(type(customMt) == "table", "customMt", customMt)
    then
        local modData = g_thModManager:getLoadedMod(modName)
        local modEnv = THUtils.getModEnvironment(modName)
        if modData == nil or modEnv == nil then
            THUtils.stackMsg("Invalid mod environment (%s)", modName)
            return
        end
        local self = setmetatable({}, customMt)
        self.modName = modData.name
        self.modPath = modData.path
        self.modData = modData
        self.dataKey = dataKey
        self.xmlBaseKey = dataKey
        self.xmlFilename = string.format("%s.xml", self.dataKey)
        self.i18n = modEnv.g_i18n
        self.messageCenter = modEnv.g_messageCenter
        self.inputManager = modEnv.g_inputBinding
        self.delayedLoadTasks = {}
        self.actionEventIds = {}
        if rootEnv == nil then
            local mpLoadingScreen = modEnv.g_mpLoadingScreen
            local missionInfo = mpLoadingScreen.missionInfo
            local missionDynamicInfo = mpLoadingScreen.missionDynamicInfo
            self.isClient = modEnv.g_client ~= nil
            self.isServer = modEnv.g_server ~= nil
            self.isMultiplayer = missionDynamicInfo.isMultiplayer
            self.rootEnvironment = self
            self.isRootEnvironment = true
            self.thModManager = g_thModManager
            self.xmlManager = modEnv.g_xmlManager
            self.taskManager = modEnv.g_asyncTaskManager
            self.fruitTypeManager = modEnv.g_fruitTypeManager
            self.fillTypeManager = modEnv.g_fillTypeManager
            self.densityMapHeightManager = modEnv.g_densityMapHeightManager
            self.farmlandManager = modEnv.g_farmlandManager
            self.fieldManager = modEnv.g_fieldManager
            self.animationManager = modEnv.g_animationManager
            self.effectManager = modEnv.g_effectManager
            self.soundManager = modEnv.g_soundManager
            self.gui = modEnv.g_gui
            self.baseHUDFilename = modEnv.g_baseHUDFilename
            self.baseUIFilename = modEnv.g_baseUIFilename
            self.modList = missionDynamicInfo.mods
            local modSettingsPath = g_modSettingsDirectory or ""
            if modSettingsPath ~= "" then
                self.modSettingsPath = modSettingsPath
                self.settingsXMLPath = string.format("%s%s/", modSettingsPath, self.modName)
                createFolder(self.settingsXMLPath)
            end
            local mapInfo = missionInfo.map
            local mapName = mapInfo.id
            if mapInfo.isModMap then
                if THUtils.getNoNil(mapInfo.customEnvironment, "") ~= "" then
                    mapName = mapInfo.customEnvironment
                end
            end
            self.mapId = mapName:upper()
            self.mapName = mapName
            self.mapPath = ""
            self.mapInfo = mapInfo
            self.mapIsMod = mapInfo.isModMap
            self.mapXMLFilename = mapInfo.mapXMLFilename
            if self.mapIsMod then
                self.mapPath = mapInfo.baseDirectory
                self.mapModData = self.thModManager:getLoadedMod(self.mapName)
            end
            self.mapSettingsModName = "FS22_TH_AAMapSettings"
            self.mapSettingsRootKey = "mapSettings"
            self.mapSettingsModData = self.thModManager:getLoadedMod(self.mapSettingsModName)
            if self.mapSettingsModData ~= nil then
                self.mapSettingsModPath = self.mapSettingsModData.path
                self.mapSettingsXMLPath = self.mapSettingsModPath .. "maps/"
                self.mapSettingsXMLFilename = self.mapSettingsXMLPath .. "mapSettings.xml"
            end
        else
            self.isClient = rootEnv.isClient
            self.isServer = rootEnv.isServer
            self.isMultiplayer = rootEnv.isMultiplayer
            self.rootEnvironment = rootEnv
            self.isRootEnvironment = false
            self.thModManager = rootEnv.thModManager
            self.thEventManager = rootEnv.thEventManager
            self.thInputManager = rootEnv.thInputManager
            self.thSpecManager = rootEnv.thSpecManager
            self.thSettingsManager = rootEnv.thSettingsManager
            self.xmlManager = rootEnv.xmlManager
            self.taskManager = rootEnv.taskManager
            self.fruitTypeManager = rootEnv.fruitTypeManager
            self.fillTypeManager = rootEnv.fillTypeManager
            self.densityMapHeightManager = rootEnv.densityMapHeightManager
            self.farmlandManager = rootEnv.farmlandManager
            self.fieldManager = rootEnv.fieldManager
            self.animationManager = rootEnv.animationManager
            self.effectManager = rootEnv.effectManager
            self.soundManager = rootEnv.soundManager
            self.gui = rootEnv.gui
            self.baseHUDFilename = rootEnv.baseHUDFilename
            self.baseUIFilename = rootEnv.baseUIFilename
            self.modList = rootEnv.modList
            self.modSettingsPath = rootEnv.modSettingsPath
            self.settingsXMLPath = rootEnv.settingsXMLPath
            self.mapId = rootEnv.mapId
            self.mapName = rootEnv.mapName
            self.mapPath = rootEnv.mapPath
            self.mapInfo = rootEnv.mapInfo
            self.mapIsMod = rootEnv.mapIsMod
            self.mapXMLFilename = rootEnv.mapXMLFilename
            self.mapModData = rootEnv.mapModData
            self.mapSettingsModName = rootEnv.mapSettingsModName
            self.mapSettingsRootKey = rootEnv.mapSettingsRootKey
            self.mapSettingsModData = rootEnv.mapSettingsModData
            self.mapSettingsModPath = rootEnv.mapSettingsModPath
            self.mapSettingsXMLPath = rootEnv.mapSettingsXMLPath
            self.mapSettingsXMLFilename = rootEnv.mapSettingsXMLFilename
            if self.dataKey ~= rootEnv.dataKey then
                self.xmlBaseKey = string.format("%s.%s", rootEnv.dataKey, self.dataKey)
                self.xmlFilename = string.format("%s_%s.xml", rootEnv.dataKey, self.dataKey)
            end
        end
        self.networkId = self:createNetworkId()
        self.isPreInitFinished = false
        self.isPostInitFinished = false
        return self
    end
end
function THAbstractSystem:loadModSettings()
    local xmlSchema = self.xmlSchemaSettings
    local xmlBaseKey = self.xmlBaseKey
    local loadSuccess = true
    if xmlSchema ~= nil and self.xmlFilename ~= nil and self.settingsXMLPath ~= nil then
        local xmlFilename = Utils.getFilename(self.xmlFilename, self.settingsXMLPath)
        local xmlFile = XMLFile.loadIfExists("THModSettingsXML", xmlFilename, xmlSchema)
        if xmlFile ~= nil then
            loadSuccess = THUtils.pcall(self.loadFromModSettingsXMLFile, self, xmlFile, xmlBaseKey)
            xmlFile:delete()
        end
    end
    return loadSuccess
end
function THAbstractSystem:loadFromModSettingsXMLFile(xmlFile, xmlKey)
end
function THAbstractSystem:saveModSettings()
    local rootEnv = self.rootEnvironment
    local xmlSchema = self.xmlSchemaSettings
    local saveSuccess = true
    if xmlSchema ~= nil and self.xmlFilename ~= nil and self.settingsXMLPath ~= nil then
        local xmlFilename = Utils.getFilename(self.xmlFilename, self.settingsXMLPath)
        local xmlFile = nil
        if fileExists(xmlFilename) then
            xmlFile = XMLFile.load("THModSettingsXML", xmlFilename, xmlSchema)
        else
            xmlFile = XMLFile.create("THModSettingsXML", xmlFilename, rootEnv.dataKey, xmlSchema)
        end
        if xmlFile ~= nil then
            saveSuccess = THUtils.pcall(self.saveToModSettingsXMLFile, self, xmlFile, self.xmlBaseKey)
            if saveSuccess then
                xmlFile:save()
            end
            xmlFile:delete()
        end
    end
    return saveSuccess
end
function THAbstractSystem:saveToModSettingsXMLFile(xmlFile, xmlKey)
end
function THAbstractSystem:getMapSettingsXMLFiles(xmlBaseKey, mapName, allowDefault)
    mapName = mapName or self.mapName
    if mapName == nil or mapName == "" then
        THUtils.warningMsg("No valid map name found")
        return
    end
    if THUtils.argIsValid(type(xmlBaseKey) == "string" and xmlBaseKey ~= "", "xmlBaseKey", xmlBaseKey)
        and THUtils.argIsValid(type(mapName) == "string", "mapName", mapName)
    then
        if self.mapSettingsModData ~= nil then
            local mapSettingsXMLFile = XMLFile.loadIfExists("THMapSettingsXML", self.mapSettingsXMLFilename)
            if mapSettingsXMLFile ~= nil then
                local mapId = mapName:upper()
                local rootKey = self.mapSettingsRootKey
                local isFilenameFound = false
                local mapXMLFilename = nil
                local defaultXMLFilename = nil
                mapSettingsXMLFile:iterate(rootKey .. ".maps.map", function(_, pMapKey)
                    local xmlMapName = mapSettingsXMLFile:getString(pMapKey .. "#name", "")
                    local xmlMapId = xmlMapName:upper()
                    if xmlMapId == mapId then
                        local mapXMLKey = string.format("%s.%s", pMapKey, xmlBaseKey)
                        if allowDefault ~= false then
                            local baseLoadDefault = mapSettingsXMLFile:getBool(pMapKey .. "#loadDefault", true)
                            allowDefault = mapSettingsXMLFile:getBool(mapXMLKey .. "#loadDefault", baseLoadDefault)
                        end
                        mapXMLFilename = mapSettingsXMLFile:getString(mapXMLKey .. "#filename")
                        if mapXMLFilename ~= nil then
                            mapXMLFilename = Utils.getFilename(mapXMLFilename, self.mapSettingsModPath)
                            isFilenameFound = true
                            if not fileExists(mapXMLFilename) then
                                THUtils.xmlMsg(mapXMLKey, "filename", mapXMLFilename, false, THXMLMessage.FILE_NOT_FOUND)
                                mapXMLFilename = nil
                            end
                        end
                    end
                    return false
                end)
                if allowDefault ~= false then
                    local defaultXMLKey = string.format("%s.default.%s", rootKey, xmlBaseKey)
                    defaultXMLFilename = mapSettingsXMLFile:getString(defaultXMLKey .. "#filename")
                    if defaultXMLFilename ~= nil then
                        defaultXMLFilename = Utils.getFilename(defaultXMLFilename, self.mapSettingsModPath)
                        isFilenameFound = true
                        if not fileExists(defaultXMLFilename) then
                            THUtils.xmlMsg(defaultXMLKey, "filename", defaultXMLFilename, false, THXMLMessage.FILE_NOT_FOUND)
                            defaultXMLFilename = nil
                        end
                    end
                end
                mapSettingsXMLFile:delete()
                return mapXMLFilename, defaultXMLFilename, isFilenameFound
            end
        end
    end
end
function THAbstractSystem:loadMapSettings(xmlBaseKey, xmlSchema, mapName, loadFunc, loadTarget, ...)
    local mapXMLFilename, defaultXMLFilename, isFilenameFound = self:getMapSettingsXMLFiles(xmlBaseKey, mapName)
    local loadSuccess = true
    if isFilenameFound then
        local defaultXMLFile = THUtils.loadXMLFile("THMapSettingsXML", defaultXMLFilename, xmlSchema)
        if defaultXMLFile ~= nil then
            THUtils.displayMsg(THMessage.LOADING, defaultXMLFilename)
            loadSuccess = THUtils.pcall(loadFunc, loadTarget, defaultXMLFile, ...)
            THUtils.deleteXMLFile(defaultXMLFile)
        end
        if loadSuccess then
            local mapXMLFile = THUtils.loadXMLFile("THMapSettingsXML", mapXMLFilename, xmlSchema)
            if mapXMLFile ~= nil then
                THUtils.displayMsg(THMessage.LOADING, mapXMLFilename)
                loadSuccess = THUtils.pcall(loadFunc, loadTarget, mapXMLFile, ...)
                THUtils.deleteXMLFile(mapXMLFile)
            end
        end
    end
    return loadSuccess
end
function THAbstractSystem:getMapXMLFile(xmlBaseKey, mapName)
    if THUtils.argIsValid(type(xmlBaseKey) == "string" and xmlBaseKey ~= "", "xmlBaseKey", xmlBaseKey)
        and THUtils.argIsValid(mapName == nil or type(mapName) == "string", "mapName", mapName)
    then
        if self.mapIsMod and self.mapModData ~= nil then
            mapName = mapName or self.mapName
            if mapName == nil or mapName == "" then
                THUtils.warningMsg("No valid map name found")
                return
            end
            local mapXMLFilename = Utils.getFilename(self.mapXMLFilename, self.mapPath)
            local mapXMLFile = XMLFile.loadIfExists("THMapXML", mapXMLFilename)
            if mapXMLFile ~= nil then
                local dataXMLKey = string.format("map.%s", xmlBaseKey)
                local dataXMLFilename = mapXMLFile:getString(dataXMLKey .. "#filename")
                local isFilenameFound = false
                if dataXMLFilename ~= nil then
                    dataXMLFilename = Utils.getFilename(dataXMLFilename, self.mapPath)
                    isFilenameFound = true
                    if not fileExists(dataXMLFilename) then
                        dataXMLFilename = nil
                        THUtils.xmlMsg(dataXMLKey, "filename", dataXMLFilename, false, THXMLMessage.FILE_NOT_FOUND)
                    end
                end
                mapXMLFile:delete()
                return dataXMLFilename, isFilenameFound
            end
        end
    end
end
function THAbstractSystem:preLoadConfiguration()
    local xmlSchema = self.xmlSchema
    local xmlBaseKey = self.xmlBaseKey
    local loadSuccess = true
    if xmlSchema ~= nil then
        local xmlFilename = self:getMapXMLFile(xmlBaseKey)
        local xmlFile = XMLFile.loadIfExists("THMapSettingsXML", xmlFilename, xmlSchema)
        if xmlFile ~= nil then
            loadSuccess = THUtils.pcall(self.preLoadFromConfigXMLFile, self, xmlFile, xmlBaseKey)
            xmlFile:delete()
        end
        if loadSuccess then
            loadSuccess = self:loadMapSettings(xmlBaseKey, xmlSchema, nil, self.preLoadFromConfigXMLFile, self, xmlBaseKey)
        end
    end
    return loadSuccess
end
function THAbstractSystem:preLoadFromConfigXMLFile(xmlFile, xmlKey)
end
function THAbstractSystem:loadConfiguration()
    local xmlSchema = self.xmlSchema
    local xmlBaseKey = self.xmlBaseKey
    local loadSuccess = true
    if xmlSchema ~= nil then
        local xmlFilename = self:getMapXMLFile(xmlBaseKey)
        local xmlFile = XMLFile.loadIfExists("THMapXML", xmlFilename, xmlSchema)
        if xmlFile ~= nil then
            loadSuccess = THUtils.pcall(self.loadFromConfigXMLFile, self, xmlFile, xmlBaseKey)
            xmlFile:delete()
        end
        if loadSuccess then
            loadSuccess = self:loadMapSettings(xmlBaseKey, xmlSchema, nil, self.loadFromConfigXMLFile, self, xmlBaseKey)
        end
    end
    return loadSuccess
end
function THAbstractSystem:loadFromConfigXMLFile(xmlFile, xmlKey)
    return true
end
function THAbstractSystem:loadFromSavegame()
    if self.isServer then
        local savegame = self.savegame
        local xmlSchema = self.xmlSchemaSavegame
        if xmlSchema ~= nil and savegame ~= nil
            and savegame.savegameDirectory ~= nil
        then
            local savegameXMLFilename = string.format("%s/%s", savegame.savegameDirectory, self.xmlFilename)
            local savegameXMLFile = XMLFile.loadIfExists("THSavegameXML", savegameXMLFilename, xmlSchema)
            if savegameXMLFile ~= nil then
                self:loadFromSavegameXMLFile(savegameXMLFile, self.xmlBaseKey)
                savegameXMLFile:delete()
            end
        end
    end
end
function THAbstractSystem:loadFromSavegameXMLFile(xmlFile, xmlKey)
    return true
end
function THAbstractSystem:saveToSavegame()
    if self.isServer then -- safety
        local rootEnv = self.rootEnvironment
        local savegame = self.savegame
        local xmlSchema = self.xmlSchemaSavegame
        if xmlSchema ~= nil and savegame ~= nil
            and savegame.savegameDirectory ~= nil
        then
            local savegameXMLFilename = string.format("%s/%s", savegame.savegameDirectory, self.xmlFilename)
            local savegameXMLFile = XMLFile.create("THSavegameXML", savegameXMLFilename, rootEnv.dataKey, xmlSchema)
            if savegameXMLFile ~= nil then
                THUtils.call(function()
                    if self:saveToSavegameXMLFile(savegameXMLFile, self.xmlBaseKey) then
                        savegameXMLFile:save()
                    end
                end)
                savegameXMLFile:delete()
            end
        end
    end
end
function THAbstractSystem:saveToSavegameXMLFile(xmlFile, xmlKey)
    return true
end
function THAbstractSystem:onPreInit()
    if not self.isPreInitFinished then
        self.isPreInitFinished = true
    end
end
function THAbstractSystem:onPostInit()
    if not self.isPostInitFinished then
        self.isPostInitFinished = true
    end
end
function THAbstractSystem:onPreCreateHUDComponents(hud, uiScale)
    if self.isRootEnvironment then
        self.hud = hud
    else
        self.hud = self.rootEnvironment.hud
    end
end
function THAbstractSystem:onHUDComponentsCreated(hud, uiScale)
end
function THAbstractSystem:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
    if self.isRootEnvironment then
        self.mission = mission
        self.missionInfo = missionInfo
        self.missionDynamicInfo = missionDynamicInfo
        self.savegameController = mission.savegameController
        self.savegameIndex = missionInfo.savegameIndex
        if self.isServer and self.savegameIndex ~= nil then
            self.savegame = self.savegameController:getSavegame(self.savegameIndex)
        end
    else
        local rootEnv = self.rootEnvironment
        self.mission = rootEnv.mission
        self.missionInfo = rootEnv.missionInfo
        self.missionDynamicInfo = rootEnv.missionDynamicInfo
        self.savegameController = rootEnv.savegameController
        self.savegameIndex = rootEnv.savegameIndex
        self.savegame = rootEnv.savegame
    end
end
function THAbstractSystem:onSendEventIds(connection)
end
function THAbstractSystem:onLoadMapFinished(mission)
end
function THAbstractSystem:onLoadEnvironment(environment, mission)
end
function THAbstractSystem:onUpdate(dt)
end
function THAbstractSystem:onUpdateTick(dt)
end
function THAbstractSystem:onRegisterActionEvents(mission)
end
function THAbstractSystem:onUnregisterActionEvents(mission)
end
function THAbstractSystem:onReadStream(streamId, connection)
    if connection:getIsServer() then
        self.networkId = streamReadUIntN(streamId, THAbstractSystem.NETWORK_ID_BITS)
        THAbstractSystem.networkObjectToId[self] = self.networkId
        if self.networkId > 0 then
            THAbstractSystem.networkObjects[self.networkId] = self
        end
    end
end
function THAbstractSystem:onWriteStream(streamId, connection)
    if not connection:getIsServer() then
        streamWriteUIntN(streamId, self.networkId, THAbstractSystem.NETWORK_ID_BITS)
    end
end
function THAbstractSystem:onReadUpdateStream(streamId, timeStamp, connection)
end
function THAbstractSystem:onWriteUpdateStream(streamId, connection, dirtyMask)
end
function THAbstractSystem:addSchemaFunctions()
    if self.isPreInitFinished then
        THUtils.stackMsg("Schema registration failed")
    else
        self.xmlManager:addCreateSchemaFunction(function()
            THUtils.call(function()
                local schemaName = string.format("%s.%s", self.modName, self.dataKey)
                self.xmlSchema = XMLSchema.new(schemaName)
                self.xmlSchemaSettings = XMLSchema.new(schemaName .. "_settings")
                self.xmlSchemaSavegame = XMLSchema.new(schemaName .. "_savegame")
            end)
        end)
        self.xmlManager:addInitSchemaFunction(function()
            self.initXMLSchema(self.xmlSchema, self.xmlBaseKey)
            self.initXMLSchemaSettings(self.xmlSchemaSettings, self.xmlBaseKey)
            self.initXMLSchemaSavegame(self.xmlSchemaSavegame, self.xmlBaseKey)
        end)
        return true
    end
    return false
end
function THAbstractSystem.getNetworkId(target)
    if target ~= nil then
        local networkId = THAbstractSystem.networkObjectToId[target]
        if networkId ~= nil and networkId > 0 then
            if THAbstractSystem.networkObjects[networkId] ~= target then
                THAbstractSystem.networkObjects[networkId] = target
                THUtils.stackMsg(THAbstractSystem.MESSAGE.NETWORK_ID_MISMATCH, networkId)
            end
            return networkId
        end
    end
    return 0
end
function THAbstractSystem.getTargetByNetworkId(networkId)
    if networkId ~= nil and networkId > 0 then
        local target = THAbstractSystem.networkObjects[networkId]
        if target ~= nil then
            if THAbstractSystem.networkObjectToId[target] ~= networkId then
                THAbstractSystem.networkObjectToId[target] = networkId
                THUtils.stackMsg(THAbstractSystem.MESSAGE.NETWORK_ID_MISMATCH, networkId)
            end
            return target
        end
    end
end
function THAbstractSystem.createNetworkId(target)
    local networkId = 0
    if g_server ~= nil
        and THUtils.argIsValid(THUtils.getIsType(target, THAbstractSystem), "target", target)
    then
        networkId = target:getNetworkId()
        if networkId == nil or networkId <= 0 then
            local nextId = #THAbstractSystem.networkObjects + 1
            local maxId = (2 ^ THAbstractSystem.NETWORK_ID_BITS) - 1
            if nextId <= maxId then
                THAbstractSystem.networkObjects[nextId] = target
                THAbstractSystem.networkObjectToId[target] = nextId
                networkId = nextId
            else
                networkId = 0
                THUtils.stackMsg(THAbstractSystem.MESSAGE.NETWORK_ID_LIMIT_REACHED, maxId)
            end
        end
    end
    return networkId
end
function THAbstractSystem:raiseActive()
    local thEventManager = self.rootEnvironment.thEventManager
    if thEventManager ~= nil then
        return thEventManager:raiseActive()
    end
end
function THAbstractSystem:getNextDirtyFlag()
    local thEventManager = self.rootEnvironment.thEventManager
    if thEventManager ~= nil then
        return thEventManager:getNextDirtyFlag()
    end
end
function THAbstractSystem:raiseDirtyFlags(dirtyFlag)
    local thEventManager = self.rootEnvironment.thEventManager
    if thEventManager ~= nil then
        return thEventManager:raiseDirtyFlags(dirtyFlag)
    end
end
function THAbstractSystem:addProtectedTask(taskFunc)
    self.taskManager:addTask(function()
        return THUtils.call(taskFunc)
    end)
end
function THAbstractSystem:addProtectedSubTask(taskFunc)
    self.taskManager:addSubtask(function()
        return THUtils.call(taskFunc)
    end)
end
function THAbstractSystem:addDelayedLoadTask(taskFunc)
    if THUtils.argIsValid(type(taskFunc) == "function", "taskFunc", taskFunc) then
        local isTaskFound = false
        for _, otherTaskFunc in pairs(self.delayedLoadTasks) do
            if taskFunc == otherTaskFunc then
                isTaskFound = true
                break
            end
        end
        if not isTaskFound then
            table.insert(self.delayedLoadTasks, taskFunc)
            return true
        end
    end
    return false
end
function THAbstractSystem:runDelayedLoadTasks()
    local numTasks = #self.delayedLoadTasks
    if numTasks > 0 then
        for taskIndex = 1, numTasks do
            THUtils.call(self.delayedLoadTasks[taskIndex])
            self.delayedLoadTasks[taskIndex] = nil
        end
    end
end
function THAbstractSystem:createDataTable(parent, useRaw, dataClass, ...)
    if THUtils.argIsValid(type(parent) == "table", "parent", parent)
        and THUtils.argIsValid(dataClass == nil
            or (type(dataClass) == "table" and dataClass.isa ~= nil), "dataClass", dataClass)
    then
        local dataTable = nil
        if useRaw == true then
            dataTable = rawget(parent, self)
        else
            dataTable = parent[self]
        end
        if dataTable == nil then
            if dataClass ~= nil then
                dataTable = dataClass.newData(parent, self, ...)
            else
                dataTable = {}
            end
            dataTable.rootEnvironment = THUtils.getNoNil(dataTable.rootEnvironment, self)
            dataTable.parent = THUtils.getNoNil(dataTable.parent, parent)
            if useRaw == true then
                rawset(parent, self, dataTable)
            else
                parent[self] = dataTable
            end
        else
            if dataClass ~= nil and (dataTable.isa == nil or not dataTable:isa(dataClass)) then
                THUtils.stackMsg("Cannot change data class of existing data table")
                return
            end
        end
        return dataTable, parent
    end
end
function THAbstractSystem:getDataTable(parent, useRaw)
    if parent ~= nil then -- nil values pass through
        local dataTable = nil
        if useRaw == true then
            dataTable = rawget(parent, self)
        else
            dataTable = parent[self]
        end
        return dataTable, parent
    end
end