-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THModManager = {}
local THModManager_mt = THUtils.createClass(THModManager)
function THModManager.new(customMt)
    customMt = customMt or THModManager_mt
    local mpLoadingScreen = g_mpLoadingScreen
    local missionDynamicInfo = mpLoadingScreen.missionDynamicInfo
    local modList = missionDynamicInfo.mods
    local self = setmetatable({}, customMt)
    self.loadedMods = {
        byId = {},
        byIndex = {},
    }
    self.modList = modList
    self:initLoadedMods()
    return self
end
function THModManager:initLoadedMods()
    for _, modInfo in ipairs(self.modList) do
        local modName = THUtils.getNoNil(modInfo.modName, "")
        if modName ~= "" then
            local modData = {
                id = modName:upper(),
                name = modName,
                info = modInfo,
                path = modInfo.modDir,
                file = modInfo.modFile,
                index = #self.loadedMods.byIndex + 1,
                title = modInfo.title,
            }
            local modVersion = modInfo.version
            if type(modVersion) == "string" then
                local versionData = THUtils.splitString(modVersion, ".")
                if versionData == nil then
                    versionData = {}
                end
                modData.version = {
                    text = modVersion,
                    major = tonumber(versionData[1] or 0) or 0,
                    minor = tonumber(versionData[2] or 0) or 0,
                    revision = tonumber(versionData[3] or 0) or 0,
                    build = tonumber(versionData[4] or 0) or 0
                }
            else
                modData.version = {
                    text = "0.0.0.0",
                    major = 0,
                    minor = 0,
                    revision = 0,
                    build = 0
                }
            end
            self.loadedMods.byId[modData.id] = modData
            self.loadedMods.byIndex[modData.index] = modData
        end
    end
end
function THModManager:getLoadedMod(modName)
    local modData = nil -- placeholder
    if type(modName) == "string" then
        modData = self.loadedMods.byId[modName:upper()]
    elseif modName ~= nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "modName", modName)
    end
    return modData
end
function THModManager:getLoadedModByClassName(className, modName)
    local modData = nil -- placeholder
    if modName ~= nil then
        modData = self:getLoadedMod(modName)
        if modData == nil then
            return
        end
    end
    if modData == nil then
        local loadedModsArray, numLoadedMods = self:getLoadedMods()
        if numLoadedMods > 0 then
            for modIndex = 1, numLoadedMods do
                modData = loadedModsArray[modIndex]
                local modClass = THUtils.getClassObject(className, modData.name)
                if type(modClass) == "table" then
                    return modData, modClass
                end
            end
        end
    else
        local modClass = THUtils.getClassObject(className, modData.name)
        if type(modClass) == "table" then
            return modData, modClass
        end
    end
end
function THModManager:getLoadedMods()
    return self.loadedMods.byIndex, #self.loadedMods.byIndex
end
function THModManager:getModVersion(modName)
    local modData = self:getLoadedMod(modName)
    local modVersionText = "0.0.0.0"
    local modVersionData = nil
    if modData ~= nil then
        modVersionData = modData.version
        if modVersionData ~= nil then
            modVersionText = modVersionData.text or modVersionText
        end
    end
    return modVersionText, modVersionData
end
_G.g_thModManager = THModManager.new()