-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THExtendedHudBox = {}
local THExtendedHudBox_mt = THUtils.createClass(THExtendedHudBox, KeyValueInfoHUDBox)
local debugFlagId = THDebugUtil.createDebugFlag("THExtendedHudBox")
THExtendedHudBox.debugFlagId = debugFlagId
THExtendedHudBox.OFFSET = {
    LABEL_TEXT = { 0, 4 },
    TITLE_SHADOW = { 2, 2 },
    TEXT_SHADOW = { 2, 2 }
}
THExtendedHudBox.SIZE = {
    DEFAULT_TEXT = { 0, 14 },
    TITLE_TEXT = { 0, 18 },
    INDICATOR = { 3, 18 },
    BACKGROUND = { 240, 0 }, -- default 340, 0
    MARGIN = { 6, 4 }, -- default 16, 15
    ROW = { 308, 20 } -- default 308, 26
}
THExtendedHudBox.COLOR = {
    BACKGROUND = { 0, 0, 0, 0 },
    TITLE_SHADOW = { 0, 0, 0, 1 },
    TEXT_SHADOW = { 0, 0, 0, 1 }
}
local function getCustomData(self)
    local customData = nil
    if self ~= nil then
        customData = rawget(self, self)
    end
    return customData
end
THExtendedHudBox.getCustomData = getCustomData
function THExtendedHudBox.new(uiScale, customMt)
    customMt = customMt or THExtendedHudBox_mt
    local self = KeyValueInfoHUDBox:superClass().new(customMt, uiScale)
    self.displayComponents = {}
    self.cachedLines = {}
    self.activeLines = {}
    self.title = ""
    local customData = {}
    customData.titleShadowOffsetX = 0
    customData.titleShadowOffsetY = 0
    customData.textShadowOffsetX = 0
    customData.textShadowOffsetY = 0
    customData.titleColor = KeyValueInfoHUDBox.COLOR.TEXT_DEFAULT
    customData.indicatorWidth = 0
    customData.indicatorHeight = 0
    customData.newActiveLines = {}
    rawset(self, self, customData)
    self:storeScaledValues()
    return self
end
function THExtendedHudBox:draw(baseX, baseY, ...)
    local customData = getCustomData(self)
    if customData ~= nil then
        local rightPosX = baseX
        local leftPosX = baseX - self.boxWidth
        local posY = baseY
        local height = (2 * self.listMarginHeight) + (#self.activeLines * self.rowHeight) -- start at the bottom and work up
        drawFilledRect(leftPosX, posY, self.boxWidth, height, THUtils.unpack(THExtendedHudBox.COLOR.BACKGROUND))
        setTextAlignment(RenderText.ALIGN_LEFT)
        setTextBold(true)
        setTextColor(THUtils.unpack(THExtendedHudBox.COLOR.TITLE_SHADOW))
        renderText((leftPosX + self.labelTextOffsetX) - customData.titleShadowOffsetX, (posY + height + self.labelTextOffsetY) - customData.titleShadowOffsetY, self.titleTextSize, self.title)
        if customData.titleColor ~= nil then
            setTextColor(THUtils.unpack(customData.titleColor))
        else
            setTextColor(THUtils.unpack(KeyValueInfoHUDBox.COLOR.TEXT_DEFAULT))
        end
        renderText(leftPosX + self.labelTextOffsetX, posY + height + self.labelTextOffsetY, self.titleTextSize, self.title)
        setTextBold(false)
        drawFilledRect(leftPosX, posY + height, self.boxWidth, 1 / g_screenHeight, unpack(KeyValueInfoHUDBox.COLOR.SEPARATOR))
        posY = posY + self.listMarginHeight
        local leftTextPosX = leftPosX + self.leftTextOffsetX + self.listMarginWidth + customData.indicatorWidth
        local rightTextPosX = rightPosX - self.rightTextOffsetX - self.listMarginWidth
        local textAreaX = self.boxWidth - self.leftTextOffsetX - self.listMarginWidth - self.rightTextOffsetX - self.listMarginWidth
        local numActiveLines = #self.activeLines
        if numActiveLines > 0 then
            for lineIndex = numActiveLines, 1, -1 do
                local lineInfo = self.activeLines[lineIndex]
                local indicatorOffsetY = self.rowHeight - customData.indicatorHeight
                if lineInfo.indicatorColor ~= nil then
                    drawFilledRect(leftPosX, posY + indicatorOffsetY, customData.indicatorWidth, customData.indicatorHeight, THUtils.unpack(lineInfo.indicatorColor))
                end
                if lineInfo.backgroundColor ~= nil then
                    local bkgPosX = leftPosX + customData.indicatorWidth
                    local bkgWidth = self.rowWidth - customData.indicatorWidth
                    drawFilledRect(bkgPosX, posY, bkgWidth, self.rowHeight, THUtils.unpack(lineInfo.backgroundColor))
                end
                setTextAlignment(RenderText.ALIGN_LEFT)
                setTextColor(THUtils.unpack(THExtendedHudBox.COLOR.TEXT_SHADOW))
                if lineInfo.accentuate then
                    setTextBold(true)
                else
                    setTextBold(false)
                end
                renderText(leftTextPosX - customData.textShadowOffsetX, (posY + self.leftTextOffsetY) - customData.textShadowOffsetY, self.rowTextSize, lineInfo.key)
                if lineInfo.accentuateColor ~= nil then
                    setTextColor(THUtils.unpack(lineInfo.accentuateColor))
                else
                    setTextColor(THUtils.unpack(KeyValueInfoHUDBox.COLOR.TEXT_DEFAULT))
                end
                renderText(leftTextPosX, posY + self.leftTextOffsetY, self.rowTextSize, lineInfo.key)
                setTextAlignment(RenderText.ALIGN_RIGHT)
                if lineInfo.accentuateRight then
                    setTextBold(true)
                else
                    setTextBold(false)
                end
                local maxWidth = textAreaX - (self.boxWidth * 0.025) - getTextWidth(self.rowTextSize, lineInfo.key)
                local rightText = Utils.limitTextToWidth(lineInfo.value, self.rowTextSize, maxWidth, false, "...")
                setTextColor(THUtils.unpack(THExtendedHudBox.COLOR.TEXT_SHADOW))
                renderText(rightTextPosX - customData.textShadowOffsetX, (posY + self.rightTextOffsetY) - customData.textShadowOffsetY, self.rowTextSize, rightText)
                if lineInfo.accentuateColorRight ~= nil then
                    setTextColor(THUtils.unpack(lineInfo.accentuateColorRight))
                else
                    setTextColor(THUtils.unpack(KeyValueInfoHUDBox.COLOR.TEXT_DEFAULT))
                end
                renderText(rightTextPosX, posY + self.rightTextOffsetY, self.rowTextSize, rightText)
                posY = posY + self.rowHeight
            end
        end
        setTextAlignment(RenderText.ALIGN_LEFT)
        self.doShowNextFrame = false
    end
end
function THExtendedHudBox:storeScaledValues(...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                local uiScale = self.uiScale
                local _ = nil -- placeholder
                customData.titleShadowOffsetX, customData.titleShadowOffsetY = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.OFFSET.TITLE_SHADOW))
                customData.textShadowOffsetX, customData.textShadowOffsetY = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.OFFSET.TEXT_SHADOW))
                _, self.labelTextOffsetY = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.OFFSET.LABEL_TEXT))
                self.listMarginWidth, self.listMarginHeight = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.SIZE.MARGIN))
                customData.indicatorWidth, customData.indicatorHeight = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.SIZE.INDICATOR))
                _, self.labelTextSize = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.SIZE.TITLE_TEXT))
                _, self.rowTextSize = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.SIZE.DEFAULT_TEXT))
                self.titleTextSize = self.labelTextSize
                self.boxWidth = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.SIZE.BACKGROUND))
                self.rowWidth, self.rowHeight = THUtils.getNormalizedValues(uiScale, THUtils.unpack(THExtendedHudBox.SIZE.ROW))
            end)
        end
        return ...
    end
    return appendFunc(THExtendedHudBox:superClass().storeScaledValues(self, ...))
end
function THExtendedHudBox:xtSetTitle(title, titleColor, ...)
    local customData = getCustomData(self)
    local function appendFunc(...)
        if customData ~= nil then
            THUtils.call(function()
                if THUtils.argIsValid(type(titleColor) == "table" and #titleColor == 4, "titleColor", titleColor) then
                    customData.titleColor = titleColor
                end
            end)
        end
        return ...
    end
    return appendFunc(THExtendedHudBox:superClass().setTitle(self, title, ...))
end
function THExtendedHudBox:xtAddLine(key, value, accentuate, accentuateColor, accentuateRight, accentuateColorRight, backgroundColor, indicatorColor, ...)
    local customData = getCustomData(self)
    local oldActiveLines = self.activeLines
    if customData ~= nil then
        THUtils.call(function()
            self.activeLines = THUtils.clearTable(customData.newActiveLines)
        end)
    end
    local function appendFunc(...)
        self.activeLines = oldActiveLines -- restore
        local newLineInfo = nil
        if customData ~= nil then
            THUtils.call(function()
                newLineInfo = customData.newActiveLines[1]
                if newLineInfo ~= nil
                    and THUtils.argIsValid(not accentuateRight or accentuateRight == true, "accentuateRight", accentuateRight)
                    and THUtils.argIsValid(accentuateColorRight == nil or (type(accentuateColorRight) == "table" and #accentuateColorRight >= 3), "accentuateColorRight", accentuateColorRight)
                    and THUtils.argIsValid(backgroundColor == nil or (type(backgroundColor) == "table" and #backgroundColor >= 3), "backgroundColor", backgroundColor)
                    and THUtils.argIsValid(indicatorColor == nil or (type(indicatorColor) == "table" and #indicatorColor >= 3), "indicatorColor", indicatorColor)
                then
                    newLineInfo.accentuateRight = THUtils.getNoNil(accentuateRight, newLineInfo.accentuate)
                    newLineInfo.accentuateColorRight = accentuateColorRight
                    newLineInfo.backgroundColor = backgroundColor
                    newLineInfo.indicatorColor = indicatorColor
                    table.insert(self.activeLines, newLineInfo)
                end
            end)
        end
        return newLineInfo, ...
    end
    return appendFunc(THExtendedHudBox:superClass().addLine(self, key, value, accentuate, accentuateColor, ...))
end