-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

THFillTypeManager = {}
local THFillTypeManager_mt = THUtils.createClass(THFillTypeManager, THAbstractSystem)
THFillTypeManager.DATA_KEY = "fillTypes"
local debugFlagId = THDebugUtil.createDebugFlag("THFillTypeManager")
THFillTypeManager.debugFlagId = debugFlagId
function THFillTypeManager.new(rootEnv, dataKey, modName, customMt)
    dataKey = dataKey or THFillTypeManager.DATA_KEY
    customMt = customMt or THFillTypeManager_mt
    local self = THAbstractSystem.new(rootEnv, dataKey, modName, customMt)
    if self ~= nil then
        if rootEnv.fillTypeManager == nil then
            THUtils.stackMsg("Can not find parent fillTypeManager")
        else
            self.parent = rootEnv.fillTypeManager
            self.fillTypes = {}
            self.thEventManager:addEventListener(self, self.modName)
            return self
        end
    end
end
function THFillTypeManager:initializeDataTables()
    local fillTypesArray, numFillTypes = self:getFillTypes()
    if numFillTypes > 0 then
        for fillTypeIndex, fillTypeInfo in ipairs(fillTypesArray) do
            if self.fillTypes[fillTypeIndex] == nil then
                local fillTypeData = {}
                fillTypeData.index = fillTypeInfo.index
                fillTypeData.name = fillTypeInfo.name
                fillTypeData.title = fillTypeInfo.title
                fillTypeData.info = fillTypeInfo
                self.fillTypes[fillTypeIndex] = fillTypeData
            end
        end
    end
    return true
end
function THFillTypeManager:finalizeDataTables()
    local fillTypesArray, numFillTypes = self:getFillTypes()
    if numFillTypes > 0 then
        for fillTypeIndex, fillTypeInfo in ipairs(fillTypesArray) do
            local fillTypeData = self:getFillTypeData(fillTypeIndex)
            if fillTypeData == nil then
                THUtils.stackMsg(THMessage.INITIALIZATION_FAILED, "fillType", fillTypeInfo.name)
            end
        end
    end
    return true
end
function THFillTypeManager:onPostInit()
    if not self.isPostInitFinished then
        THFillTypeManager:superClass().onPostInit(self)
        self:addDelayedLoadTask(function()
            THUtils.setFunctionHook("FillTypeManager", "loadMapData", THFillTypeManager, nil, false, self)
            THUtils.setFunctionHook("DensityMapHeightManager", "loadMapData", THFillTypeManager, "hook_loadDensityMapHeightData", false, self)
        end)
        self.isPostInitFinished = true
    end
end
function THFillTypeManager:onSetMissionInfo(mission, missionInfo, missionDynamicInfo)
    THFillTypeManager:superClass().onSetMissionInfo(self, mission, missionInfo, missionDynamicInfo)
end
function THFillTypeManager:onLoadMapFinished(mission)
    THFillTypeManager:superClass().onLoadMapFinished(self, mission)
    self:initializeDataTables()
    self:finalizeDataTables()
end
function THFillTypeManager:getFillType(fillType)
    local manager = self.parent
    local targetType = type(fillType)
    local fillTypeInfo = nil
    if targetType == "number" then
        fillTypeInfo = manager:getFillTypeByIndex(fillType)
    elseif targetType == "string" then
        fillTypeInfo = manager:getFillTypeByName(fillType:upper())
    elseif targetType == "table" and fillType.index ~= nil then
        fillTypeInfo = manager:getFillTypeByIndex(fillType.index)
        if fillTypeInfo ~= fillType then
            fillTypeInfo = nil
        end
    elseif fillType ~= nil then
        THUtils.stackMsg(THMessage.ARGUMENT_INVALID, "fillType", fillType)
    end
    return fillTypeInfo
end
function THFillTypeManager:getFillTypeName(fillType)
    local fillTypeInfo = self:getFillType(fillType)
    if fillTypeInfo ~= nil and fillTypeInfo.name ~= nil then
        return fillTypeInfo.name
    end
    return "UNKNOWN"
end
function THFillTypeManager:getFillTypeData(fillType)
    local fillTypeInfo = self:getFillType(fillType)
    local fillTypeData = nil
    if fillTypeInfo ~= nil then
        fillTypeData = self.fillTypes[fillTypeInfo.index]
    end
    return fillTypeData
end
function THFillTypeManager:getFillTypes()
    local manager = self.parent
    local fillTypesArray = manager:getFillTypes()
    local numFillTypes = 0
    if fillTypesArray ~= nil then
        numFillTypes = #fillTypesArray
    end
    return fillTypesArray, numFillTypes
end
function THFillTypeManager.hook_loadMapData(superFunc, self, manager, xmlFile, missionInfo, baseDirectory, ...)
    local function appendFunc(rSuccess, ...)
        if rSuccess and self.mapSettingsModData ~= nil then
            THUtils.call(function()
                self:loadMapSettings("fillTypes", nil, nil, manager.loadFillTypes, manager, self.mapSettingsModPath, false, self.mapSettingsModName)
                manager:constructFillTypeTextureArrays()
            end)
        end
        return rSuccess, ...
    end
    return appendFunc(superFunc(manager, xmlFile, missionInfo, baseDirectory, ...))
end
function THFillTypeManager.hook_loadDensityMapHeightData(superFunc, self, manager, xmlFile, missionInfo, baseDirectory, ...)
    local function appendFunc(rSuccess, ...)
        if rSuccess and self.mapSettingsModData ~= nil then
            THUtils.call(function()
                self:loadMapSettings("densityMapHeightTypes", nil, nil, manager.loadDensityMapHeightTypes, manager, missionInfo, self.mapSettingsModPath)
            end)
        end
        return rSuccess, ...
    end
    return appendFunc(superFunc(manager, xmlFile, missionInfo, baseDirectory, ...))
end