-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thAdvancedAddon.modName
local thDataKey = g_thAdvancedAddon.dataKey
THPlaceable = {}
local debugFlagId = THDebugUtil.createDebugFlag("THPlaceable")
THPlaceable.debugFlagId = debugFlagId
THPlaceable.MOD_NAME = thModName
THPlaceable.SPEC_NAME = "placeable"
THPlaceable.OBJECT_TYPE = {
    UNKNOWN = 0,
    LOADING_STATION = 1,
    UNLOADING_STATION = 2,
    STORAGE = 3
}
THPlaceable.OBJECT_TYPE_ID = {}
for typeId, typeIndex in pairs(THPlaceable.OBJECT_TYPE) do
    THPlaceable.OBJECT_TYPE_ID[typeIndex] = typeId
end
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THPlaceable.SPEC_NAME, thModName, verbose)
end
THPlaceable.getSpecTable = getSpecTable
function THPlaceable.initSpecialization()
end
function THPlaceable.prerequisitesPresent(specializations)
    return true
end
function THPlaceable.registerEvents(placeableType)
end
function THPlaceable.registerEventListeners(placeableType)
    SpecializationUtil.registerEventListener(placeableType, "onPreLoad", THPlaceable)
    SpecializationUtil.registerEventListener(placeableType, "onDelete", THPlaceable)
end
function THPlaceable.registerFunctions(placeableType)
    SpecializationUtil.registerFunction(placeableType, "thAA_getPlaceableObject", THPlaceable.thAA_getPlaceableObject)
    SpecializationUtil.registerFunction(placeableType, "thAA_getPlaceableObjectById", THPlaceable.thAA_getPlaceableObjectById)
    SpecializationUtil.registerFunction(placeableType, "thAA_getPlaceableObjects", THPlaceable.thAA_getPlaceableObjects)
    SpecializationUtil.registerFunction(placeableType, "thAA_addPlaceableObject", THPlaceable.thAA_addPlaceableObject)
    SpecializationUtil.registerFunction(placeableType, "thAA_removePlaceableObject", THPlaceable.thAA_removePlaceableObject)
    SpecializationUtil.registerFunction(placeableType, "thAA_addPlaceableObjectToFarm", THPlaceable.thAA_addPlaceableObjectToFarm)
    SpecializationUtil.registerFunction(placeableType, "thAA_removePlaceableObjectFromFarm", THPlaceable.thAA_removePlaceableObjectFromFarm)
end
function THPlaceable.registerOverwrittenFunctions(placeableType)
end
function THPlaceable:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.registeredObjects = {
        byIndex = {},
        byTarget = {},
        byType = {}
    }
    for _, typeIndex in pairs(THPlaceable.OBJECT_TYPE) do
        specTable.registeredObjects.byType[typeIndex] = {}
    end
    specTable.registeredFarmObjects = {}
    local maxNumFarms = FarmManager.MAX_NUM_FARMS or 1
    if maxNumFarms > 0 then
        for farmId = 1, maxNumFarms do
            specTable.registeredFarmObjects[farmId] = {
                byIndex = {},
                byTarget = {},
                byType = {}
            }
            for _, typeIndex in pairs(THPlaceable.OBJECT_TYPE) do
                specTable.registeredFarmObjects[farmId].byType[typeIndex] = {}
            end
        end
    end
    specTable.functionParameters = {
        placeable = self
    }
    if not specTable.thIsInitialized then
        THUtils.setFunctionHook("Storage", "load", nil, THPlaceable.inj_storageLoad, false, specTable.functionParameters)
        THUtils.setFunctionHook("LoadingStation", "load", nil, THPlaceable.inj_stationLoad, false, specTable.functionParameters)
        THUtils.setFunctionHook("UnloadingStation", "load", nil, THPlaceable.inj_stationLoad, false, specTable.functionParameters)
        THUtils.setFunctionHook("Storage", "delete", nil, THPlaceable.inj_deleteObject, false, specTable.functionParameters)
        THUtils.setFunctionHook("LoadingStation", "delete", nil, THPlaceable.inj_deleteObject, false, specTable.functionParameters)
        THUtils.setFunctionHook("UnloadingStation", "delete", nil, THPlaceable.inj_deleteObject, false, specTable.functionParameters)
        THUtils.setFunctionHook("Storage", "setOwnerFarmId", nil, THPlaceable.inj_setObjectOwnerFarmId, false, specTable.functionParameters)
        THUtils.setFunctionHook("LoadingStation", "setOwnerFarmId", nil, THPlaceable.inj_setObjectOwnerFarmId, false, specTable.functionParameters)
        THUtils.setFunctionHook("UnloadingStation", "setOwnerFarmId", nil, THPlaceable.inj_setObjectOwnerFarmId, false, specTable.functionParameters)
        THUtils.setFunctionHook("StorageSystem", "addStorage", nil, THPlaceable.inj_addStorage, false, specTable.functionParameters)
        THUtils.setFunctionHook("StorageSystem", "removeStorage", nil, THPlaceable.inj_removeStorage, false, specTable.functionParameters)
        THUtils.setFunctionHook("StorageSystem", "addLoadingStation", nil, THPlaceable.inj_addStation, false, specTable.functionParameters)
        THUtils.setFunctionHook("StorageSystem", "removeLoadingStation", nil, THPlaceable.inj_removeStation, false, specTable.functionParameters)
        THUtils.setFunctionHook("StorageSystem", "addUnloadingStation", nil, THPlaceable.inj_addStation, false, specTable.functionParameters)
        THUtils.setFunctionHook("StorageSystem", "removeUnloadingStation", nil, THPlaceable.inj_removeStation, false, specTable.functionParameters)
        specTable.thIsInitialized = true
    end
end
function THPlaceable:onDelete()
    local specTable = getSpecTable(self, false)
    if specTable ~= nil then
        THUtils.clearTable(specTable.functionParameters)
    end
end
function THPlaceable:thAA_getPlaceableObject(object, farmId)
    local specTable = getSpecTable(self)
    if object ~= nil then
        if farmId == nil or farmId == AccessHandler.EVERYONE then
            return specTable.registeredObjects.byTarget[object]
        elseif farmId ~= AccessHandler.NOBODY
            and specTable.registeredFarmObjects[farmId] ~= nil
        then
            return specTable.registeredFarmObjects[farmId].byTarget[object]
        end
    end
end
function THPlaceable:thAA_getPlaceableObjectById(objectId, objectType, farmId)
    local idType = type(objectId)
    if idType == "number" or idType == "string" then
        local objectsArray, numObjects = self:thAA_getPlaceableObjects(objectType, farmId)
        if objectsArray ~= nil and numObjects > 0 then
            if idType == "string" then
                objectId = objectId:upper()
                for objectIndex = 1, numObjects do
                    local object = objectsArray[objectIndex]
                    local objectData = THPlaceable.getObjectData(object)
                    if objectData ~= nil and objectData.id == objectId then
                        return object
                    end
                end
            else
                return objectsArray[objectId]
            end
        end
    elseif objectId ~= nil then
        THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "objectId", objectId)
    end
end
function THPlaceable:thAA_getPlaceableObjects(objectType, farmId)
    local specTable = getSpecTable(self)
    if objectType ~= nil and THPlaceable.OBJECT_TYPE_ID[objectType] == nil then
        THPlaceable.errorMsg(self, true, THMessage.ARGUMENT_INVALID, "objectType", objectType)
    else
        local objectsArray = nil
        if farmId == nil or farmId == AccessHandler.EVERYONE then
            if objectType == nil then
                objectsArray = specTable.registeredObjects.byIndex
            else
                objectsArray = specTable.registeredObjects.byType[objectType]
            end
        elseif farmId ~= AccessHandler.NOBODY
            and specTable.registeredFarmObjects[farmId] ~= nil
        then
            if objectType == nil then
                objectsArray = specTable.registeredFarmObjects[farmId].byIndex
            else
                objectsArray = specTable.registeredFarmObjects[farmId].byType[objectType]
            end
        end
        if objectsArray ~= nil then
            return objectsArray, #objectsArray
        end
    end
    return nil, 0
end
function THPlaceable:thAA_addPlaceableObject(object)
    local specTable = getSpecTable(self)
    local objectData = THPlaceable.getObjectData(object)
    local isObjectAdded = false
    if objectData ~= nil and objectData.loadingPlaceable == self then
        if specTable.registeredObjects.byTarget[object] == nil then
            table.insert(specTable.registeredObjects.byIndex, object)
            specTable.registeredObjects.byTarget[object] = object
            if objectData.type ~= nil then
                table.insert(specTable.registeredObjects.byType[objectData.type], object)
            end
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THPlaceable.displayMsg(self, "Added object:")
                THUtils.printTable(objectData, 1)
                THPlaceable.displayMsg(self, "- Objects table:")
                THUtils.printTable(specTable.registeredObjects, 2)
            end
            isObjectAdded = true
        end
    end
    return isObjectAdded
end
function THPlaceable:thAA_removePlaceableObject(object)
    local specTable = getSpecTable(self)
    local isObjectRemoved = false
    if object ~= nil and self:thAA_getPlaceableObject(object) then
        local objectData = THPlaceable.getObjectData(object)
        local objectIndex = 1
        self:thAA_removePlaceableObjectFromFarm(object)
        while true do
            if specTable.registeredObjects.byIndex[objectIndex] == nil then
                break
            end
            if specTable.registeredObjects.byIndex[objectIndex] == object then
                table.remove(specTable.registeredObjects.byIndex, objectIndex)
                isObjectRemoved = true
            else
                objectIndex = objectIndex + 1
            end
        end
        if objectData ~= nil and objectData.type ~= nil
            and specTable.registeredObjects.byType[objectData.type] ~= nil
        then
            local typedObjectsArray = specTable.registeredObjects.byType[objectData.type]
            objectIndex = 1
            while true do
                if typedObjectsArray[objectIndex] == nil then
                    break
                end
                if typedObjectsArray[objectIndex] == object then
                    table.remove(typedObjectsArray, objectIndex)
                    isObjectRemoved = true
                else
                    objectIndex = objectIndex + 1
                end
            end
        end
        if specTable.registeredObjects.byTarget[object] ~= nil then
            specTable.registeredObjects.byTarget[object] = nil
            isObjectRemoved = true
        end
        if isObjectRemoved then
            if THDebugUtil.getIsEnabled(debugFlagId) then
                THPlaceable.displayMsg(self, "Removed object:")
                if objectData ~= nil then
                    THUtils.printTable(objectData, 1)
                else
                    THUtils.printTable(object)
                end
                THPlaceable.displayMsg(self, "- Objects table:")
                THUtils.printTable(specTable.registeredObjects, 2)
            end
        end
    end
    return isObjectRemoved
end
function THPlaceable:thAA_addPlaceableObjectToFarm(object, farmId)
    local specTable = getSpecTable(self)
    local objectData = THPlaceable.getObjectData(object)
    local isObjectAdded = false
    if objectData ~= nil and objectData.loadingPlaceable == self
        and self:thAA_getPlaceableObject(object)
    then
        local function addObjectToFarm(pFarmId)
            local objectsTable = specTable.registeredFarmObjects[pFarmId]
            if objectsTable ~= nil then
                if objectsTable.byTarget[object] == nil then
                    table.insert(objectsTable.byIndex, object)
                    objectsTable.byTarget[object] = object
                    if objectData.type ~= nil then
                        table.insert(objectsTable.byType[objectData.type], object)
                    end
                    if THDebugUtil.getIsEnabled(debugFlagId) then
                        THPlaceable.displayMsg(self, "Added object to farm %s:", pFarmId)
                        if objectData ~= nil then
                            THUtils.printTable(objectData)
                        else
                            THUtils.printTable(object)
                        end
                    end
                    isObjectAdded = true
                end
            end
        end
        if farmId == nil or farmId == AccessHandler.EVERYONE then
            for otherFarmId in pairs(specTable.registeredFarmObjects) do
                addObjectToFarm(otherFarmId)
            end
        elseif farmId ~= AccessHandler.NOBODY then
            addObjectToFarm(farmId)
        end
    end
    return isObjectAdded
end
function THPlaceable:thAA_removePlaceableObjectFromFarm(object, farmId)
    local specTable = getSpecTable(self)
    local isObjectRemoved = false
    if object ~= nil and self:thAA_getPlaceableObject(object) then
        local objectData = THPlaceable.getObjectData(object)
        local function removeObjectFromFarm(pFarmId)
            local objectsTable = specTable.registeredFarmObjects[pFarmId]
            if objectsTable ~= nil then
                local objectIndex = 1
                local isRemoved = false
                while true do
                    if objectsTable.byIndex[objectIndex] == nil then
                        break
                    end
                    if objectsTable.byIndex[objectIndex] == object then
                        table.remove(objectsTable.byIndex, objectIndex)
                        isRemoved = true
                    else
                        objectIndex = objectIndex + 1
                    end
                end
                if objectData ~= nil and objectData.type ~= nil
                    and objectsTable.byType[objectData.type] ~= nil
                then
                    local typedObjectsArray = objectsTable.byType[objectData.type]
                    objectIndex = 1
                    while true do
                        if typedObjectsArray[objectIndex] == nil then
                            break
                        end
                        if typedObjectsArray[objectIndex] == object then
                            table.remove(typedObjectsArray, objectIndex)
                            isRemoved = true
                        else
                            objectIndex = objectIndex + 1
                        end
                    end
                end
                if objectsTable.byTarget[object] ~= nil then
                    objectsTable.byTarget[object] = nil
                    isRemoved = true
                end
                if isRemoved then
                    if THDebugUtil.getIsEnabled(debugFlagId) then
                        THPlaceable.displayMsg(self, "Removed object from farm %s", pFarmId)
                        if objectData ~= nil then
                            THUtils.printTable(objectData)
                        else
                            THUtils.printTable(object)
                        end
                    end
                    isObjectRemoved = true
                end
            end
        end
        if farmId == nil or farmId == AccessHandler.EVERYONE then
            for otherFarmId in pairs(specTable.registeredFarmObjects) do
                removeObjectFromFarm(otherFarmId)
            end
        elseif farmId ~= AccessHandler.NOBODY then
            removeObjectFromFarm(farmId)
        end
    end
    return isObjectRemoved
end
function THPlaceable.getName(self)
    if self ~= nil then
        local name = Placeable.getName(self)
        if name ~= nil and name ~= "" then
            return name
        end
    end
    return "Unknown"
end
function THPlaceable.displayMsg(self, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local placeableName = nil
        if self ~= nil then
            placeableName = THPlaceable.getName(self)
        end
        if placeableName ~= nil and text ~= nil then
            outputText = string.format("%s >> %s", placeableName, text)
        end
    end
    THUtils.displayMsg(outputText, ...)
end
function THPlaceable.warningMsg(self, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local placeableName = nil
        if self ~= nil then
            placeableName = THPlaceable.getName(self)
        end
        if placeableName ~= nil and text ~= nil then
            outputText = string.format("%s >> %s", placeableName, text)
        end
    end
    THUtils.warningMsg(outputText, ...)
end
function THPlaceable.errorMsg(self, showStack, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local placeableName = nil
        if self ~= nil then
            placeableName = THPlaceable.getName(self)
        end
        if placeableName ~= nil and text ~= nil then
            outputText = string.format("%s >> %s", placeableName, text)
        end
    end
    THUtils.errorMsg(showStack, outputText, ...)
end
function THPlaceable.debugMsg(self, flagId, level, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local placeableName = nil
        if self ~= nil then
            placeableName = THPlaceable.getName(self)
        end
        if placeableName ~= nil then
            outputText = string.format("%s >> %s", placeableName, text)
        end
    end
    THDebugUtil.printMsg(flagId, level, outputText, ...)
end
function THPlaceable.xmlMsg(self, xmlKey, attrName, attrValue, isError, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local placeableName = nil
        if self ~= nil then
            placeableName = THPlaceable.getName(self)
        end
        if placeableName ~= nil and text ~= nil then
            outputText = string.format("%s >> %s", placeableName, text)
        end
    end
    return THUtils.xmlMsg(xmlKey, attrName, attrValue, isError, outputText, ...)
end
function THPlaceable.getObjectData(object)
    if object ~= nil then
        return g_thAdvancedAddon:getDataTable(object)
    end
end
function THPlaceable.getStationData(station)
    return THPlaceable.getObjectData(station)
end
function THPlaceable.getConfigSchemaPath(specClass)
    if THUtils.argIsValid(type(specClass) == "table", "specClass", specClass) then
        return string.format("placeable.%s.%s", thDataKey, specClass.SPEC_NAME)
    end
    return ""
end
function THPlaceable.getConfigBaseKey(specClass)
    if THUtils.argIsValid(type(specClass) == "table", "specClass", specClass) then
        return string.format("placeable.%s.%s", thDataKey, specClass.SPEC_NAME)
    end
    return ""
end
function THPlaceable.getSavegameSchemaPath(specClass)
    if THUtils.argIsValid(type(specClass) == "table", "specClass", specClass) then
        return string.format("placeables.placeable(?).%s.%s", specClass.MOD_NAME, specClass.SPEC_NAME)
    end
    return ""
end
function THPlaceable.getSavegameBaseKey(specClass, savegame)
    if THUtils.argIsValid(savegame ~= nil and savegame.key ~= nil, "savegame", savegame)
        and THUtils.argIsValid(type(specClass) == "table", "specClass", specClass)
    then
        return string.format("%s.%s.%s", savegame.key, specClass.MOD_NAME, specClass.SPEC_NAME)
    end
    return ""
end
function THPlaceable.inj_storageLoad(superFunc, extraArgs, storage, components, xmlFile, xmlKey, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and storage ~= nil and xmlFile == self.xmlFile
        and storage.isa ~= nil and storage:isa(Storage)
    then
        THUtils.call(function()
            local storageData = THPlaceable.getObjectData(storage)
            local objectType = THPlaceable.OBJECT_TYPE.STORAGE
            local objectTypeId = THPlaceable.OBJECT_TYPE_ID[objectType]
            if storageData == nil then
                storageData = g_thAdvancedAddon:createDataTable(storage)
            end
            if storageData == nil then
                THPlaceable.xmlMsg(self, xmlKey, nil, nil, true, "could not create storage data table")
                return
            end
            local otherStorage = self:thAA_getPlaceableObject(storage)
            if otherStorage == nil or otherStorage == storage then
                local xmlBaseKey = string.format("%s.%s", xmlKey, thDataKey)
                local storageId = xmlFile:getString(xmlBaseKey .. "#id", storageData.id or "")
                if storageId == "" then
                    local nextStorageIndex = #specTable.registeredObjects.byType[objectType] + 1
                    storageId = string.format("%s_%d", objectTypeId, nextStorageIndex)
                end
                storageData.id = storageId:upper()
                storageData.type = objectType
                storageData.xmlKey = xmlKey
                storageData.loadingPlaceable = self
                if otherStorage == nil then
                    if not self:thAA_addPlaceableObject(storage) then
                        THPlaceable.xmlMsg(self, xmlKey, nil, nil, false, "Could not add storage %q to placeable", storageData.id)
                    end
                end
            else
                THPlaceable.xmlMsg(self, xmlKey, nil, nil, true, "Storage tables are corrupt (storage mismatch)")
            end
        end)
    end
    return superFunc(storage, components, xmlFile, xmlKey, ...)
end
function THPlaceable.inj_stationLoad(superFunc, extraArgs, station, components, xmlFile, xmlKey, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and station ~= nil and xmlFile == self.xmlFile
        and station.isa ~= nil
    then
        THUtils.call(function()
            local stationData = THPlaceable.getObjectData(station)
            local objectType = THPlaceable.OBJECT_TYPE.UNKNOWN
            if stationData == nil then
                stationData = g_thAdvancedAddon:createDataTable(station)
                stationData.linkedStorageIds = {}
            end
            if stationData == nil then
                THPlaceable.xmlMsg(self, xmlKey, nil, nil, true, "could not create station data table")
                return
            end
            if station:isa(LoadingStation) then
                objectType = THPlaceable.OBJECT_TYPE.LOADING_STATION
            elseif station:isa(UnloadingStation) then
                objectType = THPlaceable.OBJECT_TYPE.UNLOADING_STATION
            end
            if objectType == THPlaceable.OBJECT_TYPE.UNKNOWN then
                THPlaceable.xmlMsg(self, xmlKey, nil, nil, false, "cannot determine station type")
            end
            local otherStation = self:thAA_getPlaceableObject(station)
            if otherStation == nil or otherStation == station then
                local xmlBaseKey = string.format("%s.%s", xmlKey, thDataKey)
                local objectTypeId = THPlaceable.OBJECT_TYPE_ID[objectType]
                local stationId = xmlFile:getString(xmlBaseKey .. "#id", stationData.id or "")
                local storageIdList = xmlFile:getString(xmlBaseKey .. ".storageIds")
                if stationId == "" then
                    local nextStationIndex = #specTable.registeredObjects.byType[objectType] + 1
                    stationId = string.format("%s_%d", objectTypeId, nextStationIndex)
                end
                stationData.id = stationId:upper()
                stationData.type = objectType
                stationData.xmlKey = xmlKey
                stationData.loadingPlaceable = self
                if storageIdList ~= nil then
                    THUtils.clearTable(stationData.linkedStorageIds)
                    if storageIdList ~= "" then
                        local storageIdsArray = THUtils.splitString(storageIdList, " ")
                        if storageIdsArray ~= nil then
                            for _, storageId in ipairs(storageIdsArray) do
                                if storageId ~= "" then
                                    stationData.linkedStorageIds[storageId:upper()] = true
                                end
                            end
                        end
                    end
                end
                if otherStation == nil then
                    if not self:thAA_addPlaceableObject(station) then
                        THPlaceable.xmlMsg(self, xmlKey, nil, nil, false, "Could not add station %q to placeable", stationData.id)
                    end
                end
            else
                THPlaceable.xmlMsg(self, xmlKey, nil, nil, true, "station tables are corrupt (station mismatch")
            end
        end)
    end
    return superFunc(station, components, xmlFile, xmlKey, ...)
end
function THPlaceable.inj_deleteObject(superFunc, extraArgs, object, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and object ~= nil then
        local function appendFunc(...)
            THUtils.call(function()
                self:thAA_removePlaceableObject(object)
            end)
            return ...
        end
        return appendFunc(superFunc(object, ...))
    end
    return superFunc(object, ...)
end
function THPlaceable.inj_setObjectOwnerFarmId(superFunc, extraArgs, object, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and object ~= nil then
        local function appendFunc(...)
            THUtils.call(function()
                self:thAA_removePlaceableObjectFromFarm(object)
                self:thAA_addPlaceableObjectToFarm(object, object.ownerFarmId)
            end)
            return ...
        end
        return appendFunc(superFunc(object, ...))
    end
    return superFunc(object, ...)
end
function THPlaceable.inj_addStorage(superFunc, extraArgs, storageSystem, storage, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and storage ~= nil then
        local function appendFunc(rSuccess, ...)
            if rSuccess == true then
                THUtils.call(function()
                    self:thAA_removePlaceableObjectFromFarm(storage)
                    self:thAA_addPlaceableObjectToFarm(storage, storage.ownerFarmId)
                end)
            end
            return rSuccess, ...
        end
        return appendFunc(superFunc(storageSystem, storage, ...))
    end
    return superFunc(storageSystem, storage, ...)
end
function THPlaceable.inj_removeStorage(superFunc, extraArgs, storageSystem, storage, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and storage ~= nil then
        local function appendFunc(rSuccess, ...)
            if rSuccess == true then
                THUtils.call(function()
                    self:thAA_removePlaceableObjectFromFarm(storage)
                end)
            end
            return rSuccess, ...
        end
        return appendFunc(superFunc(storageSystem, storage, ...))
    end
    return superFunc(storageSystem, storage, ...)
end
function THPlaceable.inj_addStation(superFunc, extraArgs, storageSystem, station, placeable, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and station ~= nil and (placeable == nil or placeable == self) then
        local function appendFunc(rSuccess, ...)
            if rSuccess == true then
                THUtils.call(function()
                    self:thAA_removePlaceableObjectFromFarm(station)
                    self:thAA_addPlaceableObjectToFarm(station, station.ownerFarmId)
                end)
            end
            return rSuccess, ...
        end
        return appendFunc(superFunc(storageSystem, station, placeable, ...))
    end
    return superFunc(storageSystem, station, placeable, ...)
end
function THPlaceable.inj_removeStation(superFunc, extraArgs, storageSystem, station, placeable, ...)
    local self = extraArgs.placeable
    local specTable = getSpecTable(self, false)
    if specTable ~= nil and station ~= nil and (placeable == nil or placeable == self) then
        local function appendFunc(rSuccess, ...)
            if rSuccess == true then
                THUtils.call(function()
                    self:thAA_removePlaceableObjectFromFarm(station)
                end)
            end
            return rSuccess, ...
        end
        return appendFunc(superFunc(storageSystem, station, placeable, ...))
    end
    return superFunc(storageSystem, station, placeable, ...)
end