-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thAdvancedAddon.modName
local thPFManager = g_thAdvancedAddon.thPFManager
local thFarmlandManager = g_thAdvancedAddon.farmlandManager
THCutter = {}
local debugFlagId = THDebugUtil.createDebugFlag("THCutter")
THCutter.debugFlagId = debugFlagId
THCutter.MOD_NAME = thModName
THCutter.SPEC_NAME = "cutter"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THCutter.SPEC_NAME, thModName, verbose)
end
THCutter.getSpecTable = getSpecTable
function THCutter.initSpecialization()
end
function THCutter.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(THVehicle, specializations)
        and SpecializationUtil.hasSpecialization(Cutter, specializations)
end
function THCutter.registerEvents(vehicleType)
end
function THCutter.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THCutter)
end
function THCutter.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "processCutterArea", THCutter.processCutterArea)
end
function THCutter:onPreLoad(savegame)
    local specTable = getSpecTable(self)
end
function THCutter:processCutterArea(superFunc, workArea, dt, ...)
    local specTable = getSpecTable(self)
    local cutterSpec = self.spec_cutter
    local workParams = cutterSpec.workAreaParameters
    specTable.lastTouchedFarmland = nil
    specTable.lastTouchedSoilType = nil
    local function appendFunc(...)
        THUtils.call(function()
            local inputFruitType = workParams.lastFruitType
            if workArea ~= nil
                and inputFruitType ~= nil and inputFruitType ~= FruitType.UNKNOWN
            then
                local wx, _, wz = getWorldTranslation(workArea.width)
                local hx, _, hz = getWorldTranslation(workArea.height)
                local mx = (wx + hx) / 2
                local mz = (wz + hz) / 2
                specTable.lastTouchedFarmland = thFarmlandManager:getFarmlandIdAtWorldPosition(mx, mz)
                specTable.lastTouchedSoilType = thPFManager:getSoilTypeIndexAtWorldPos(mx, mz, true) -- force
            end
        end)
        return ...
    end
    return appendFunc(superFunc(self, workArea, dt, ...))
end
function THCutter:hook_onEndWorkAreaProcessing(superFunc, dt, hasProcessed, ...)
    local specTable = getSpecTable(self, false)
    local cutterSpec = self.spec_cutter
    local workParams = cutterSpec.workAreaParameters
    if specTable ~= nil and workParams ~= nil then
        THUtils.call(function()
            local combineVehicle = workParams.combineVehicle
            local thCombineSpec = THCombine.getSpecTable(combineVehicle, false)
            if thCombineSpec ~= nil then
                thCombineSpec.lastValidCutterVehicle = self
            end
            specTable.lastStrawArea = workParams.lastRealArea
        end)
    end
    return superFunc(self, dt, hasProcessed, ...)
end
g_thAdvancedAddon:addDelayedLoadTask(function()
    THUtils.setFunctionHook("Cutter", "onEndWorkAreaProcessing", THCutter)
end)