-- Copyright 2024 by Todd Hundersmarck (ThundR)
-- All Rights Reserved

local thModName = g_thAdvancedAddon.modName
local thDataKey = g_thAdvancedAddon.dataKey
THVehicle = {}
local debugFlagId = THDebugUtil.createDebugFlag("THVehicle")
THVehicle.debugFlagId = debugFlagId
THVehicle.MOD_NAME = thModName
THVehicle.SPEC_NAME = "vehicle"
local function getSpecTable(self, verbose)
    return THUtils.getSpecTable(self, THVehicle.SPEC_NAME, thModName, verbose)
end
THVehicle.getSpecTable = getSpecTable
function THVehicle.initSpecialization()
end
function THVehicle.prerequisitesPresent(specializations)
    return true
end
function THVehicle.registerEvents(vehicleType)
end
function THVehicle.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onPreLoad", THVehicle)
end
function THVehicle.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "thAA_callChainFunc", THVehicle.thAA_callChainFunc)
end
function THVehicle.registerOverwrittenFunctions(vehicleType)
end
function THVehicle:onPreLoad(savegame)
    local specTable = getSpecTable(self)
    specTable.vehicleCache = {}
end
function THVehicle:thAA_callChainFunc(func, ...)
    local specTable = getSpecTable(self)
    local rootVehicle = self
    if self.getRootVehicle ~= nil then
        rootVehicle = self:getRootVehicle()
    end
    THUtils.clearTable(specTable.vehicleCache)
    THUtils.call(func, rootVehicle, ...)
    specTable.vehicleCache[rootVehicle] = true
    if rootVehicle.getChildVehicles ~= nil then
        local childVehicles = rootVehicle:getChildVehicles()
        if childVehicles ~= nil then
            for _, childVehicle in pairs(childVehicles) do
                if not specTable.vehicleCache[childVehicle] then
                    THUtils.call(func, childVehicle, ...)
                    specTable.vehicleCache[childVehicle] = true
                end
            end
        end
    end
end
function THVehicle.getName(self)
    if self ~= nil then
        local name = Vehicle.getName(self)
        if name ~= nil and name ~= "" then
            return name
        end
    end
    return "Unknown"
end
function THVehicle.displayMsg(self, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local vehicleName = nil
        if self ~= nil then
            vehicleName = THVehicle.getName(self)
        end
        if vehicleName ~= nil then
            outputText = string.format("%s >> %s", vehicleName, text)
        end
    end
    THUtils.displayMsg(outputText, ...)
end
function THVehicle.warningMsg(self, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local vehicleName = nil
        if self ~= nil then
            vehicleName = THVehicle.getName(self)
        end
        if vehicleName ~= nil then
            outputText = string.format("%s >> %s", vehicleName, text)
        end
    end
    THUtils.warningMsg(outputText, ...)
end
function THVehicle.errorMsg(self, showStack, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local vehicleName = nil
        if self ~= nil then
            vehicleName = THVehicle.getName(self)
        end
        if vehicleName ~= nil then
            outputText = string.format("%s >> %s", vehicleName, text)
        end
    end
    THUtils.errorMsg(showStack, outputText, ...)
end
function THVehicle.debugMsg(self, flagId, level, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local vehicleName = nil
        if self ~= nil then
            vehicleName = THVehicle.getName(self)
        end
        if vehicleName ~= nil then
            outputText = string.format("%s >> %s", vehicleName, text)
        end
    end
    THDebugUtil.printMsg(flagId, level, outputText, ...)
end
function THVehicle.xmlMsg(self, xmlKey, attrName, attrValue, isError, text, ...)
    local outputText = text
    if text ~= nil and text ~= "" then
        local vehicleName = nil
        if self ~= nil then
            vehicleName = THVehicle.getName(self)
        end
        if vehicleName ~= nil then
            outputText = string.format("%s >> %s", vehicleName, text)
        end
    end
    return THUtils.xmlMsg(xmlKey, attrName, attrValue, isError, outputText, ...)
end
function THVehicle.getConfigSchemaPath(specClass)
    if THUtils.argIsValid(type(specClass) == "table", "specClass", specClass) then
        return string.format("vehicle.%s.%s", thDataKey, specClass.SPEC_NAME)
    end
    return ""
end
function THVehicle.getConfigBaseKey(specClass)
    if THUtils.argIsValid(type(specClass) == "table", "specClass", specClass) then
        return string.format("vehicle.%s.%s", thDataKey, specClass.SPEC_NAME)
    end
    return ""
end
function THVehicle.getSavegameSchemaPath(specClass)
    if THUtils.argIsValid(type(specClass) == "table", "specClass", specClass) then
        return string.format("vehicles.vehicle(?).%s.%s", specClass.MOD_NAME, specClass.SPEC_NAME)
    end
    return ""
end
function THVehicle.getSavegameBaseKey(specClass, savegame)
    if THUtils.argIsValid(savegame ~= nil and savegame.key ~= nil, "savegame", savegame)
        and THUtils.argIsValid(type(specClass) == "table", "specClass", specClass)
    then
        return string.format("%s.%s.%s", savegame.key, specClass.MOD_NAME, specClass.SPEC_NAME)
    end
    return ""
end